import hashlib
import os
import pathlib
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import dpdispatcher  # noqa: F401
from dpdispatcher.base_context import BaseContext  # noqa: F401
from dpdispatcher.contexts.hdfs_context import HDFSContext  # noqa: F401
from dpdispatcher.contexts.lazy_local_context import LazyLocalContext  # noqa: F401
from dpdispatcher.contexts.local_context import LocalContext  # noqa: F401
from dpdispatcher.contexts.ssh_context import SSHContext, SSHSession  # noqa: F401

# test backward compatibility with dflow
from dpdispatcher.dpcloudserver.client import RequestInfoException as _  # noqa: F401
from dpdispatcher.entrypoints.run import run  # noqa: F401
from dpdispatcher.entrypoints.submission import handle_submission  # noqa: F401
from dpdispatcher.machine import Machine  # noqa: F401
from dpdispatcher.machines.distributed_shell import DistributedShell  # noqa: F401
from dpdispatcher.machines.dp_cloud_server import Lebesgue  # noqa: F401
from dpdispatcher.machines.JH_UniScheduler import JH_UniScheduler  # noqa: F401
from dpdispatcher.machines.lsf import LSF  # noqa: F401
from dpdispatcher.machines.pbs import PBS  # noqa: F401
from dpdispatcher.machines.shell import Shell  # noqa: F401
from dpdispatcher.machines.slurm import Slurm  # noqa: F401
from dpdispatcher.submission import Job, Resources, Submission, Task  # noqa: F401
from dpdispatcher.utils.hdfs_cli import HDFS  # noqa: F401
from dpdispatcher.utils.job_status import JobStatus  # noqa: F401
from dpdispatcher.utils.record import record  # noqa: F401
from dpdispatcher.utils.utils import RetrySignal, retry  # noqa: F401


def setUpModule():
    os.chdir(os.path.abspath(os.path.dirname(__file__)))


def get_file_md5(file_path):
    return hashlib.md5(pathlib.Path(file_path).read_bytes()).hexdigest()



import json
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
# from .context import dpdispatcher
# from dpdispatcher.local_context import LocalContext
from dpdispatcher.machine import Machine
from dpdispatcher.submission import Resources, Submission, Task

# from dpdispatcher.ssh_context import SSHContext


# from dpdispatcher.submission import
# from dpdispatcher.slurm import Slurm

# local_session = LocalSession({'work_path':'temp2'})
# local_context = LocalContext(local_root='test_slurm_dir/', work_profile=local_session)
# lazy_local_context = LazyLocalContext(local_root='test_slurm_dir/')


# machine_dict = dict(hostname='localhost', remote_root='/home/dp/dpdispatcher/tests/temp2', username='dp')
# ssh_session = SSHSession(**machine_dict)
# ssh_session = SSHSession(hostname='8.131.233.55', remote_root='/home/dp/dp_remote', username='dp')
# ssh_context = SSHContext(local_root='test_slurm_dir', ssh_session=ssh_session)
# slurm = Slurm(context=ssh_context)
# slurm = Slurm(context=lazy_local_context)

# resources = Resources(number_node=1, cpu_per_node=4, gpu_per_node=2, queue_name="GPU_2080Ti", group_size=4,
#     custom_flags=['#SBATCH --exclude=2080ti000,2080ti001,2080ti002,2080ti004,2080ti005,2080ti006'],
#     para_deg=2,
#     strategy={"if_cuda_multi_devices":True})
# slurm_sbatch_dict={'mem': '10G', 'cpus_per_task':1, 'time': "120:0:0"}
# slurm_resources = SlurmResources(resources=resources, slurm_sbatch_dict=slurm_sbatch_dict)

with open("jsons/machine_slurm.json") as f:
    mdata = json.load(f)

machine = Machine.load_from_dict(mdata["machine"])
resources = Resources.load_from_dict(mdata["resources"])

submission = Submission(
    work_base="0_md/",
    machine=machine,
    resources=resources,
    forward_common_files=["graph.pb"],
    backward_common_files=[],
)  # ,  batch=PBS)
task1 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-1/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task2 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-2/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task3 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-3/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task4 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-4/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
submission.register_task_list(
    [
        task1,
        task2,
        task3,
        task4,
    ]
)
submission.run_submission(clean=True)



# /// script
# # dpdispatcher doesn't use `requires-python` and `dependencies`
# requires-python = ">=3"
# dependencies = [
# ]
# [tool.dpdispatcher]
# work_base = "./"
# forward_common_files=[]
# backward_common_files=[]
# [tool.dpdispatcher.machine]
# batch_type = "Shell"
# local_root = "./"
# context_type = "LazyLocalContext"
# [tool.dpdispatcher.resources]
# number_node = 1
# cpu_per_node = 1
# gpu_per_node = 0
# group_size = 0
# [[tool.dpdispatcher.task_list]]
# # no need to contain the script filename
# command = "python"
# # can be a glob pattern
# task_work_path = "./"
# forward_files = []
# backward_files = ["log"]
# ///

print("hello world!")



import os
import shutil
import tarfile
from glob import glob

from dpdispatcher.base_context import BaseContext
from dpdispatcher.dlog import dlog
from dpdispatcher.utils.hdfs_cli import HDFS


class HDFSContext(BaseContext):
    def __init__(
        self,
        local_root,
        remote_root,
        remote_profile={},
        *args,
        **kwargs,
    ):
        assert isinstance(local_root, str)
        self.init_local_root = local_root
        self.init_remote_root = remote_root
        self.temp_local_root = os.path.abspath(local_root)
        self.temp_remote_root = remote_root
        self.remote_profile = remote_profile

    @classmethod
    def load_from_dict(cls, context_dict):
        local_root = context_dict["local_root"]
        remote_root = context_dict["remote_root"]
        remote_profile = context_dict.get("remote_profile", {})
        instance = cls(
            local_root=local_root,
            remote_root=remote_root,
            remote_profile=remote_profile,
        )
        return instance

    def get_job_root(self):
        return self.remote_root

    def bind_submission(self, submission):
        self.submission = submission
        self.local_root = os.path.join(self.temp_local_root, submission.work_base)
        self.remote_root = os.path.join(
            self.temp_remote_root, submission.submission_hash
        )

        HDFS.mkdir(self.remote_root)

    def _put_files(self, files, dereference=True):
        of = self.submission.submission_hash + "_upload.tgz"
        # local tar
        if os.path.isfile(os.path.join(self.local_root, of)):
            os.remove(os.path.join(self.local_root, of))
        with tarfile.open(
            os.path.join(self.local_root, of), "w:gz", dereference=dereference
        ) as tar:
            for ii in files:
                ii_full = os.path.join(self.local_root, ii)
                tar.add(ii_full, arcname=ii)

        # trans
        from_f = os.path.join(self.local_root, of)
        HDFS.copy_from_local(os.path.join(self.local_root, of), self.remote_root)

        # clean up
        os.remove(from_f)

    def upload(self, submission, dereference=True):
        """Upload forward files and forward command files to HDFS root dir.

        Parameters
        ----------
        submission : Submission class instance
            represents a collection of tasks, such as forward file names
        dereference : bool
            whether to dereference symbolic links

        Returns
        -------
        none
        """
        file_list = []

        for task in submission.belonging_tasks:
            local_job = os.path.join(self.local_root, task.task_work_path)
            for ff in task.forward_files:
                abs_file_list = glob(os.path.join(local_job, ff))
                if not abs_file_list:
                    raise FileNotFoundError(
                        "cannot find upload file " + os.path.join(local_job, ff)
                    )
                rel_file_list = [
                    os.path.relpath(ii, self.local_root) for ii in abs_file_list
                ]
                file_list.extend(rel_file_list)

        local_job = self.local_root
        for fc in submission.forward_common_files:
            abs_file_list = glob(os.path.join(local_job, fc))
            if not abs_file_list:
                raise FileNotFoundError(
                    "cannot find upload file " + os.path.join(local_job, fc)
                )
            rel_file_list = [
                os.path.relpath(ii, self.local_root) for ii in abs_file_list
            ]
            file_list.extend(rel_file_list)

        self._put_files(file_list, dereference=dereference)

    def download(
        self, submission, check_exists=False, mark_failure=True, back_error=False
    ):
        """Download backward files from HDFS root dir.

        Parameters
        ----------
        submission : Submission class instance
            represents a collection of tasks, such as backward file names
        check_exists : bool
            whether to check if the file exists
        mark_failure : bool
            whether to mark the task as failed if the file does not exist
        back_error : bool
            whether to download error files

        Returns
        -------
        none
        """
        cwd = os.getcwd()

        # download all hdfs files to tmp dir
        gz_dir = os.path.join(self.local_root, "tmp")
        if os.path.exists(gz_dir):
            shutil.rmtree(gz_dir, ignore_errors=True)
        os.mkdir(os.path.join(self.local_root, "tmp"))
        rfile_tgz = f"{self.remote_root}/{submission.submission_hash}_*_download.tar.gz"
        lfile_tgz = f"{self.local_root}/tmp/"
        HDFS.copy_to_local(rfile_tgz, lfile_tgz)

        tgz_file_list = glob(os.path.join(self.local_root, "tmp/*_download.tar.gz"))
        for tgz in tgz_file_list:
            with tarfile.open(tgz, "r:gz") as tar:
                tar.extractall(path=gz_dir)

        for task in submission.belonging_tasks:
            local_job = os.path.join(self.local_root, task.task_work_path)
            remote_job = os.path.join(gz_dir, task.task_work_path)
            flist = task.backward_files
            if back_error:
                errors = glob(os.path.join(remote_job, "error*"))
                flist.extend(errors)
            for jj in flist:
                rfile = os.path.join(remote_job, jj)
                lfile = os.path.join(local_job, jj)

                if not os.path.exists(rfile):
                    if check_exists:
                        if mark_failure:
                            with open(
                                os.path.join(
                                    self.local_root,
                                    task.task_work_path,
                                    f"tag_failure_download_{jj}",
                                ),
                                "w",
                            ) as fp:
                                pass
                        else:
                            raise FileNotFoundError(
                                "do not find download file " + rfile
                            )
                    else:
                        raise FileNotFoundError("do not find download file " + rfile)
                else:
                    if os.path.exists(lfile):
                        dlog.info(f"find existing {lfile}, replacing by {rfile}")
                        if os.path.isdir(lfile):
                            shutil.rmtree(lfile, ignore_errors=True)
                        elif os.path.isfile(lfile):
                            os.remove(lfile)
                    shutil.move(rfile, lfile)

        local_job = self.local_root
        remote_job = gz_dir
        flist = submission.backward_common_files
        if back_error:
            errors = glob(os.path.join(remote_job, "error*"))
            flist.extend(errors)
        for jj in flist:
            rfile = os.path.join(remote_job, jj)
            lfile = os.path.join(local_job, jj)

            if not os.path.exists(rfile):
                if check_exists:
                    if mark_failure:
                        with open(
                            os.path.join(self.local_root, f"tag_failure_download_{jj}"),
                            "w",
                        ) as fp:
                            pass
                    else:
                        raise FileNotFoundError("do not find download file " + rfile)
                else:
                    raise FileNotFoundError("do not find download file " + rfile)
            else:
                if os.path.exists(lfile):
                    dlog.info(f"find existing {lfile}, replacing by {rfile}")
                    if os.path.isdir(lfile):
                        shutil.rmtree(lfile, ignore_errors=True)
                    elif os.path.isfile(lfile):
                        os.remove(lfile)
                shutil.move(rfile, lfile)

        # remove tmp dir
        shutil.rmtree(gz_dir, ignore_errors=True)

    def check_file_exists(self, fname):
        """Check whether the given file exists, often used in checking whether the belonging job has finished.

        Parameters
        ----------
        fname : string
            file name to be checked

        Returns
        -------
        status: boolean
        """
        return HDFS.exists(os.path.join(self.remote_root, fname))

    def clean(self):
        HDFS.remove(self.remote_root)

    def write_file(self, fname, write_str):
        local_file = os.path.join("/tmp/", fname)
        with open(local_file, "w") as fp:
            fp.write(write_str)
        HDFS.copy_from_local(local_file, os.path.join(self.remote_root, fname))
        return local_file

    def read_file(self, fname):
        return HDFS.read_hdfs_file(os.path.join(self.remote_root, fname))

    def block_call(self, cmd):
        raise RuntimeError(
            "Unsupported method. You may use an unsupported combination of the machine and the context."
        )



import os

HTTP_TIME_OUT = 30

API_HOST = os.environ.get("DPDISPATCHER_LEBESGUE_API_HOST", "https://bohrium.dp.tech/")
API_LOGGER_STACK_INFO = os.environ.get("API_LOGGER_STACK_INFO", "")
ALI_STS_ENDPOINT = os.environ.get(
    "DPDISPATCHER_LEBESGUE_ALI_STS_ENDPOINT", "http://oss-cn-shenzhen.aliyuncs.com"
)
ALI_STS_BUCKET_NAME = os.environ.get(
    "DPDISPATCHER_LEBESGUE_ALI_STS_BUCKET_NAME", "dpcloudserver"
)
ALI_OSS_BUCKET_URL = os.environ.get(
    "DPDISPATCHER_LEBESGUE_ALI_OSS_BUCKET_URL",
    "https://dpcloudserver.oss-cn-shenzhen.aliyuncs.com/",
)



import shlex
from typing import List

from dargs import Argument

from dpdispatcher.dlog import dlog
from dpdispatcher.machine import Machine
from dpdispatcher.utils.job_status import JobStatus
from dpdispatcher.utils.utils import customized_script_header_template

pbs_script_header_template = """
#!/bin/bash -l
{select_node_line}
#PBS -j oe
{queue_name_line}
"""


class PBS(Machine):
    def gen_script(self, job):
        pbs_script = super().gen_script(job)
        return pbs_script

    def gen_script_header(self, job):
        resources = job.resources
        pbs_script_header_dict = {}
        pbs_script_header_dict["select_node_line"] = (
            f"#PBS -l select={resources.number_node}:ncpus={resources.cpu_per_node}"
        )
        if resources.gpu_per_node != 0:
            pbs_script_header_dict["select_node_line"] += (
                f":ngpus={resources.gpu_per_node}"
            )
        pbs_script_header_dict["queue_name_line"] = f"#PBS -q {resources.queue_name}"
        if (
            resources["strategy"].get("customized_script_header_template_file")
            is not None
        ):
            pbs_script_header = customized_script_header_template(
                resources["strategy"]["customized_script_header_template_file"],
                resources,
            )
        else:
            pbs_script_header = pbs_script_header_template.format(
                **pbs_script_header_dict
            )
        return pbs_script_header

    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + "_job_id"
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        self.context.write_file(fname=script_file_name, write_str=script_str)
        script_run_str = self.gen_script_command(job)
        script_run_file_name = f"{job.script_file_name}.run"
        self.context.write_file(fname=script_run_file_name, write_str=script_run_str)
        # self.context.write_file(fname=os.path.join(self.context.submission.work_base, script_file_name), write_str=script_str)
        # script_file_dir = os.path.join(self.context.submission.work_base)
        script_file_dir = self.context.remote_root
        # stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (self.context.remote_root, 'qsub', script_file_name))
        stdin, stdout, stderr = self.context.block_checkcall(
            "cd {} && {} {}".format(
                shlex.quote(script_file_dir), "qsub", shlex.quote(script_file_name)
            )
        )
        subret = stdout.readlines()
        job_id = subret[0].split()[0]
        self.context.write_file(job_id_name, job_id)
        return job_id

    def default_resources(self, resources):
        pass

    def check_status(self, job):
        job_id = job.job_id
        if job_id == "":
            return JobStatus.unsubmitted
        command = "qstat -x " + job_id
        ret, stdin, stdout, stderr = self.context.block_call(command)
        err_str = stderr.read().decode("utf-8")
        if ret != 0:
            if "qstat: Unknown Job Id" in err_str or "Job has finished" in err_str:
                if self.check_finish_tag(job=job):
                    return JobStatus.finished
                else:
                    return JobStatus.terminated
            else:
                raise RuntimeError(
                    "status command %s fails to execute. erro info: %s return code %d"
                    % (command, err_str, ret)
                )
        status_line = stdout.read().decode("utf-8").split("\n")[-2]
        status_word = status_line.split()[-2]
        # dlog.info (status_word)
        if status_word in ["Q", "H"]:
            return JobStatus.waiting
        elif status_word in ["R"]:
            return JobStatus.running
        elif status_word in ["C", "E", "K", "F"]:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else:
                return JobStatus.terminated
        else:
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        return self.context.check_file_exists(job_tag_finished)

    def kill(self, job):
        """Kill the job.

        Parameters
        ----------
        job : Job
            job
        """
        job_id = job.job_id
        ret, stdin, stdout, stderr = self.context.block_call("qdel " + str(job_id))


class Torque(PBS):
    def check_status(self, job):
        job_id = job.job_id
        if job_id == "":
            return JobStatus.unsubmitted
        command = "qstat -l " + job_id
        ret, stdin, stdout, stderr = self.context.block_call(command)
        err_str = stderr.read().decode("utf-8")
        if ret != 0:
            if "qstat: Unknown Job Id" in err_str or "Job has finished" in err_str:
                if self.check_finish_tag(job=job):
                    return JobStatus.finished
                else:
                    return JobStatus.terminated
            else:
                raise RuntimeError(
                    "status command %s fails to execute. erro info: %s return code %d"
                    % (command, err_str, ret)
                )
        status_line = stdout.read().decode("utf-8").split("\n")[-2]
        status_word = status_line.split()[-2]
        # dlog.info (status_word)
        if status_word in ["Q", "H"]:
            return JobStatus.waiting
        elif status_word in ["R"]:
            return JobStatus.running
        elif status_word in ["C", "E", "K", "F"]:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else:
                return JobStatus.terminated
        else:
            return JobStatus.unknown

    def gen_script_header(self, job):
        # ref: https://support.adaptivecomputing.com/wp-content/uploads/2021/02/torque/torque.htm#topics/torque/2-jobs/requestingRes.htm
        resources = job.resources
        pbs_script_header_dict = {}
        pbs_script_header_dict["select_node_line"] = (
            f"#PBS -l nodes={resources.number_node}:ppn={resources.cpu_per_node}"
        )
        if resources.gpu_per_node != 0:
            pbs_script_header_dict["select_node_line"] += (
                f":gpus={resources.gpu_per_node}"
            )
        pbs_script_header_dict["queue_name_line"] = f"#PBS -q {resources.queue_name}"
        if (
            resources["strategy"].get("customized_script_header_template_file")
            is not None
        ):
            pbs_script_header = customized_script_header_template(
                resources["strategy"]["customized_script_header_template_file"],
                resources,
            )
        else:
            pbs_script_header = pbs_script_header_template.format(
                **pbs_script_header_dict
            )
        return pbs_script_header


sge_script_header_template = """
#!/bin/bash
#$ -S /bin/bash
#$ -cwd
{select_node_line}
"""


class SGE(PBS):
    def __init__(
        self,
        batch_type=None,
        context_type=None,
        local_root=None,
        remote_root=None,
        remote_profile={},
        *,
        context=None,
    ):
        super(PBS, self).__init__(
            batch_type,
            context_type,
            local_root,
            remote_root,
            remote_profile,
            context=context,
        )

    def gen_script_header(self, job):
        ### Ref:https://softpanorama.org/HPC/PBS_and_derivatives/Reference/pbs_command_vs_sge_commands.shtml
        # resources.number_node is not used in SGE
        resources = job.resources
        job_name = resources.kwargs.get("job_name", "wDPjob")
        pe_name = resources.kwargs.get("pe_name", "mpi")
        sge_script_header_dict = {}
        sge_script_header_dict["select_node_line"] = f"#$ -N {job_name}\n"
        sge_script_header_dict["select_node_line"] += (
            f"#$ -pe {pe_name} {resources.cpu_per_node}\n"
        )

        if resources.queue_name != "":
            sge_script_header_dict["select_node_line"] += (
                f"#$ -q {resources.queue_name}"
            )
        if (
            resources["strategy"].get("customized_script_header_template_file")
            is not None
        ):
            file_name = resources["strategy"]["customized_script_header_template_file"]
            sge_script_header = customized_script_header_template(file_name, resources)
        else:
            sge_script_header = sge_script_header_template.format(
                **sge_script_header_dict
            )
        return sge_script_header

    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + "_job_id"
        self.context.write_file(fname=script_file_name, write_str=script_str)
        script_run_str = self.gen_script_command(job)
        script_run_file_name = f"{job.script_file_name}.run"
        self.context.write_file(fname=script_run_file_name, write_str=script_run_str)
        script_file_dir = self.context.remote_root
        stdin, stdout, stderr = self.context.block_checkcall(
            "cd {} && {} {}".format(script_file_dir, "qsub", script_file_name)
        )
        subret = stdout.readlines()
        job_id = subret[0].split()[2]
        self.context.write_file(job_id_name, job_id)
        return job_id

    def default_resources(self, resources):
        pass

    def check_status(self, job):
        job_id = job.job_id
        status_line = None
        if job_id == "":
            return JobStatus.unsubmitted
        command = "qstat"
        ret, stdin, stdout, stderr = self.context.block_call(command)
        err_str = stderr.read().decode("utf-8")
        if ret != 0:
            raise RuntimeError(
                f"status command {command} fails to execute. erro info: {err_str} return code {ret}"
            )
        status_text_list = stdout.read().decode("utf-8").split("\n")
        for txt in status_text_list:
            if job_id in txt:
                status_line = txt

        if status_line is None:
            count = 0
            while count <= 6:
                if self.check_finish_tag(job=job):
                    return JobStatus.finished
                dlog.info(
                    f"not tag_finished detected, execute sync command and wait. count {count}"
                )
                self.context.block_call("sync")
                import time

                time.sleep(10)
                count += 1
            return JobStatus.terminated
        else:
            status_word = status_line.split()[4]
            # dlog.info (status_word)
            if status_word in ["qw"]:
                return JobStatus.waiting
            elif status_word in ["r"]:
                return JobStatus.running
            else:
                return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        return self.context.check_file_exists(job_tag_finished)

    @classmethod
    def resources_subfields(cls) -> List[Argument]:
        """Generate the resources subfields.

            pe_name : str
        The parallel environment name of SGE.

        Returns
        -------
        list[Argument]
            resources subfields
        """
        doc_pe_name = "The parallel environment name of SGE system."
        doc_job_name = "The name of SGE's job."

        return [
            Argument(
                "kwargs",
                dict,
                [
                    Argument(
                        "pe_name",
                        str,
                        optional=True,
                        default="mpi",
                        doc=doc_pe_name,
                        alias=["sge_pe_name"],
                    ),
                    Argument(
                        "job_name",
                        str,
                        optional=True,
                        default="wDPjob",
                        doc=doc_job_name,
                    ),
                ],
                optional=False,
                doc="Extra arguments.",
            )
        ]



import json
import os
import sys
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
# from .context import LocalContext

# from .context import Dispatcher
from .context import (
    Task,
    setUpModule,  # noqa: F401
)
from .sample_class import SampleClass


class TestTask(unittest.TestCase):
    def setUp(self):
        self.task = SampleClass.get_sample_task()
        self.task_dict = SampleClass.get_sample_task_dict()

    def test_serialize(self):
        self.assertEqual(self.task.serialize(), self.task_dict)

    def test_deserialize(self):
        task = Task.deserialize(task_dict=self.task_dict)
        self.assertTrue(task, self.task)

    def test_serialize_deserialize(self):
        self.assertEqual(Task.deserialize(task_dict=self.task.serialize()), self.task)

    def test_task_json(self):
        with open("jsons/task.json") as f:
            task_json_dict = json.load(f)
        self.assertTrue(task_json_dict, self.task_dict)
        self.assertTrue(task_json_dict, self.task.serialize())

    def test_repr(self):
        task_repr = repr(self.task)
        print("debug:", task_repr, self.task_dict)
        self.assertEqual(task_repr, str(self.task_dict))



import logging
import os
import sys
import warnings

dlog = logging.getLogger("dpdispatcher")
dlog.propagate = False
dlog.setLevel(logging.INFO)
cwd_logfile_path = os.path.join(os.getcwd(), "dpdispatcher.log")
dlogf = logging.FileHandler(cwd_logfile_path, delay=True)
try:
    dlog.addHandler(dlogf)
    dlog.info(f"LOG INIT:dpdispatcher log direct to {cwd_logfile_path}")
except PermissionError:
    dlog.removeHandler(dlogf)
    warnings.warn(
        f"dump logfile dpdispatcher.log to {cwd_logfile_path} meet permission error. redirect the log to ~/dpdispatcher.log"
    )
    dlogf = logging.FileHandler(
        os.path.join(os.path.expanduser("~"), "dpdispatcher.log"), delay=True
    )
    dlog.addHandler(dlogf)
    dlog.info("LOG INIT:dpdispatcher log init at ~/dpdispatcher.log")

dlogf_formatter = logging.Formatter("%(asctime)s - %(levelname)s : %(message)s")
dlogf.setFormatter(dlogf_formatter)
# dlog.addHandler(dlogf)

dlog_stdout = logging.StreamHandler(sys.stdout)
dlog_stdout.setFormatter(dlogf_formatter)
dlog.addHandler(dlog_stdout)

__all__ = [
    "dlog",
]



import json
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

# from dpdispatcher.batch_object import BatchObject
from dpdispatcher.machine import Machine
from dpdispatcher.submission import Resources, Submission, Task

# from dpdispatcher.slurm import SlurmResources, Slurm

# local_session = LocalSession({'work_path':'temp2'})
# local_context = LocalContext(local_root='test_slurm_dir/', work_profile=local_session)
# lazy_local_context = LazyLocalContext(local_root='./')


# machine_dict = dict(hostname='localhost', remote_root='/home/dp/dpdispatcher/tests/temp2', username='dp')
# ssh_session = SSHSession(**machine_dict)
# ssh_session = SSHSession(hostname='8.131.233.55', remote_root='/home/dp/dp_remote', username='dp')
# ssh_context = SSHContext(local_root='test_slurm_dir', ssh_session=ssh_session)
# slurm = Slurm(context=ssh_context)
# slurm = Slurm(context=lazy_local_context)

# resources = Resources(number_node=1, cpu_per_node=4, gpu_per_node=0, queue_name="1 * NVIDIA P100", group_size=4)
# slurm_sbatch_dict={'mem': '10G', 'cpus_per_task':1, 'time': "120:0:0"}
# slurm_resources = SlurmResources(resources=resources, slurm_sbatch_dict=slurm_sbatch_dict)


# dp_cloud_server_context = DpCloudServerContext(
#     local_root='test_context_dir/',
#     username='yfb222333',
#     password='yfb222333')
# dp_cloud_server = DpCloudServer(context=dp_cloud_server_context)
# with open('test_dp_cloud_server.json', 'r') as f:
#     jdata = json.load(f)
with open("jsons/machine_dp_cloud_server.json") as f:
    compute_dict = json.load(f)

machine = Machine.load_from_dict(compute_dict["machine"])
resources = Resources.load_from_dict(compute_dict["resources"])

task1 = Task(
    command="lmp    -i input.lammps",
    task_work_path="bct-1/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task2 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-2/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task3 = Task(
    command="lmp   -i input.lammps",
    task_work_path="bct-3/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task4 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-4/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task_list = [
    task1,
    task2,
    task3,
    task4,
]

submission = Submission(
    work_base="0_md/",
    machine=machine,
    resources=resources,
    forward_common_files=["graph.pb"],
    backward_common_files=[],
    task_list=task_list,
)

submission.run_submission()



# %%
import hashlib
import os
import shutil
import sys
import unittest
import uuid
from unittest.mock import MagicMock

# %%
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
# from .context import LocalSession
from .context import (
    LocalContext,
    get_file_md5,
    setUpModule,  # noqa: F401
)

# from .context import dpd


def _identical_files(fname0, fname1):
    with open(fname0) as fp:
        code0 = hashlib.sha1(fp.read().encode("utf-8")).hexdigest()
    with open(fname1) as fp:
        code1 = hashlib.sha1(fp.read().encode("utf-8")).hexdigest()
    return code0 == code1


class TestIdFile(unittest.TestCase):
    def test_id(self):
        with open("f0", "w") as fp:
            fp.write("foo")
        with open("f1", "w") as fp:
            fp.write("foo")
        self.assertTrue(_identical_files("f0", "f1"))
        os.remove("f0")
        os.remove("f1")

    def test_diff(self):
        with open("f0", "w") as fp:
            fp.write("foo")
        with open("f1", "w") as fp:
            fp.write("bar")
        self.assertFalse(_identical_files("f0", "f1"))
        os.remove("f0")
        os.remove("f1")


class TestLocalContext(unittest.TestCase):
    def setUp(self):
        self.tmp_local_root = "test_context_dir/"
        self.tmp_remote_root = "tmp_local_context_remote_root/"
        self.local_context = LocalContext(
            local_root=self.tmp_local_root, remote_root=self.tmp_remote_root
        )

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("tmp_local_context_remote_root/")

    def test_upload_non_exist(self):
        submission_hash = "mock_hash_1"
        task1 = MagicMock(task_work_path="bct-1/", forward_files=["foo.py"])
        submission = MagicMock(
            work_base="0_md/", belonging_tasks=[task1], submission_hash=submission_hash
        )

        self.local_context.bind_submission(submission)

        with self.assertRaises(FileNotFoundError):
            self.local_context.upload(submission)

    def test_upload(self):
        submission_hash = "mock_hash_2"
        task1 = MagicMock(
            task_work_path="bct-1/",
            forward_files=["input.lammps", "conf.lmp", "some_dir/some_file"],
        )
        task2 = MagicMock(
            task_work_path="bct-2/", forward_files=["input.lammps", "conf.lmp"]
        )
        submission = MagicMock(
            work_base="0_md/",
            belonging_tasks=[task1, task2],
            forward_common_files=["graph.pb", "some_dir/some_file"],
            submission_hash=submission_hash,
        )

        self.local_context.bind_submission(submission)
        self.local_context.upload(submission)

        check_file_list = [
            "bct-1/input.lammps",
            "bct-1/conf.lmp",
            "bct-1/some_dir/some_file",
            "bct-2/input.lammps",
            "bct-2/conf.lmp",
            "graph.pb",
            "some_dir/some_file",
        ]
        for file in check_file_list:
            f1 = os.path.join(self.tmp_local_root, "0_md/", file)
            f2 = os.path.join(self.tmp_remote_root, submission_hash, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1, f2))

    # TODO: support other platforms
    @unittest.skipIf(sys.platform != "linux", "not linux")
    def test_block_call(self):
        submission_hash = "mock_hash_3"
        task1 = MagicMock(
            task_work_path="bct-1/", forward_files=["input.lammps", "conf.lmp"]
        )
        task2 = MagicMock(
            task_work_path="bct-2/", forward_files=["input.lammps", "conf.lmp"]
        )
        submission = MagicMock(
            work_base="0_md/",
            belonging_tasks=[task1, task2],
            submission_hash=submission_hash,
        )
        self.local_context.bind_submission(submission)
        self.local_context.upload(submission)
        # time.sleep(0.01)
        code, stdin, stdout, stderr = self.local_context.block_call("ls")
        self.assertEqual(code, 0)
        self.assertEqual(stdout.readlines(), ["bct-1\n", "bct-2\n"])

        code, stdin, stdout, stderr = self.local_context.block_call("ls a")
        self.assertEqual(code, 2)
        err_msg = stderr.read().decode("utf-8")
        self.assertTrue("ls: cannot access" in err_msg)
        self.assertTrue("No such file or directory\n" in err_msg)

    @unittest.skipIf(sys.platform == "win32", "sleep is not supported on Windows")
    def test_call(self):
        submission_hash = "mock_hash_4"
        submission = MagicMock(
            work_base="0_md/", belonging_tasks=[], submission_hash=submission_hash
        )
        self.local_context.bind_submission(submission)
        self.local_context.upload(submission)

        proc = self.local_context.call("sleep 0.12")
        # in some environment it's not easy to determine which is faster..
        # self.assertFalse(self.local_context.check_finish(proc))
        # time.sleep(0.06)
        # self.assertFalse(self.local_context.check_finish(proc))
        # time.sleep(0.10)
        # wait until terminated, as in some test environments, it's slow to execute commands
        proc.wait(timeout=2)
        self.assertTrue(self.local_context.check_finish(proc))
        r, o, e = self.local_context.get_return(proc)
        self.assertEqual(r, 0)
        self.assertEqual(o.read(), b"")
        self.assertEqual(e.read(), b"")
        # not correct on centos7 aliyun ehpc
        # r,o,e=self.local_context.get_return(proc)
        # self.assertEqual(r, 0)
        # self.assertEqual(o, None)
        # self.assertEqual(e, None)

    def test_file(self):
        submission_hash = "mock_hash_5"
        submission = MagicMock(
            work_base="0_md/", belonging_tasks=[], submission_hash=submission_hash
        )
        self.local_context.bind_submission(submission)
        self.local_context.upload(submission)

        self.assertFalse(self.local_context.check_file_exists("aaa"))
        tmp = str(uuid.uuid4())
        self.local_context.write_file("aaa", tmp)
        self.assertTrue(self.local_context.check_file_exists("aaa"))
        tmp1 = self.local_context.read_file("aaa")
        self.assertEqual(tmp, tmp1)


class TestLocalContextDownload(unittest.TestCase):
    # @classmethod
    # def setUpClass(cls):
    def setUp(self):
        shutil.copytree(src="test_context_dir/", dst="tmp_local_context_download_dir/")
        os.makedirs("tmp_local_context_backfill_dir/0_md/bct-1/")
        os.makedirs("tmp_local_context_backfill_dir/0_md/bct-2/")

        self.tmp_local_root = "tmp_local_context_backfill_dir"
        self.tmp_remote_root = "tmp_local_context_download_dir"
        self.local_context = LocalContext(
            local_root=self.tmp_local_root, remote_root=self.tmp_remote_root
        )

    def tearDown(self):
        shutil.rmtree("tmp_local_context_download_dir/")
        shutil.rmtree("tmp_local_context_backfill_dir/")

    def test_download_trival(self):
        # submission_hash = 'mock_hash_2'
        task1 = MagicMock(
            task_work_path="bct-1/", backward_files=["input.lammps", "conf.lmp"]
        )
        task2 = MagicMock(
            task_work_path="bct-2/", backward_files=["input.lammps", "conf.lmp"]
        )
        submission = MagicMock(
            work_base="0_md/",
            belonging_tasks=[task1, task2],
            backward_common_files=["graph.pb"],
            submission_hash="0_md/",
        )

        self.local_context.bind_submission(submission)

        check_file_list = [
            "bct-1/input.lammps",
            "bct-1/conf.lmp",
            "bct-2/input.lammps",
            "bct-2/conf.lmp",
            "graph.pb",
        ]

        self.local_context.download(submission)

        for file in check_file_list:
            f1 = os.path.join(self.tmp_local_root, "0_md/", file)
            self.assertTrue(os.path.isfile(f1))
            self.assertFalse(os.path.islink(f1))

    def test_download_check_exists(self):
        task1 = MagicMock(task_work_path="bct-1/", backward_files=["foo.py"])
        submission = MagicMock(
            work_base="0_md/",
            belonging_tasks=[task1],
            backward_common_files=["graph.pb"],
            submission_hash="0_md/",
        )
        self.local_context.bind_submission(submission)
        with self.assertRaises(FileNotFoundError):
            self.local_context.download(submission, check_exists=False)

    def test_download_mark_failure_tag(self):
        task1 = MagicMock(task_work_path="bct-1/", backward_files=["foo.py"])
        submission = MagicMock(
            work_base="0_md/",
            belonging_tasks=[task1],
            backward_common_files=["graph.pb"],
            submission_hash="0_md/",
        )
        self.local_context.bind_submission(submission)
        # with self.assertRaises(RuntimeError):
        self.local_context.download(submission, check_exists=True, mark_failure=True)

        tag_file = os.path.join(
            self.tmp_local_root, "0_md/", "bct-1/", "tag_failure_download_foo.py"
        )
        self.assertTrue(os.path.isfile(tag_file))

    def test_download_replace_old_files(self):
        task1 = MagicMock(task_work_path="bct-1/", backward_files=["input.lammps"])
        submission = MagicMock(
            work_base="0_md/",
            belonging_tasks=[task1],
            backward_common_files=["graph.pb"],
            submission_hash="0_md/",
        )
        # with self.assertRaises(RuntimeError):
        target_file = os.path.join(
            self.tmp_local_root, "0_md/", "bct-1/", "input.lammps"
        )
        with open(target_file, "w") as f:
            f.write("\n")
        md5_old = get_file_md5(target_file)
        self.local_context.bind_submission(submission)
        self.local_context.download(submission)
        md5_new = get_file_md5(target_file)
        self.assertNotEqual(md5_old, md5_new)

    def test_download_symlink(self):
        task1 = MagicMock(
            task_work_path="bct-1/", backward_files=["input.lammps.symlink"]
        )
        submission = MagicMock(
            work_base="0_md/",
            belonging_tasks=[task1],
            backward_common_files=["graph.pb.symlink"],
            submission_hash="0_md/",
        )
        os.symlink(
            os.path.abspath(
                os.path.join(self.tmp_remote_root, "0_md", "bct-1", "input.lammps")
            ),
            os.path.join(self.tmp_remote_root, "0_md", "bct-1", "input.lammps.symlink"),
        )
        os.symlink(
            os.path.abspath(os.path.join(self.tmp_remote_root, "0_md", "graph.pb")),
            os.path.join(self.tmp_remote_root, "0_md", "graph.pb.symlink"),
        )

        self.local_context.bind_submission(submission)
        self.local_context.download(submission)
        self.local_context.clean()
        task_file = os.path.join(
            self.tmp_local_root,
            "0_md",
            "bct-1",
            "input.lammps.symlink",
        )
        common_file = os.path.join(
            self.tmp_local_root,
            "0_md",
            "graph.pb.symlink",
        )
        self.assertTrue(os.path.isfile(task_file))
        self.assertTrue(os.path.isfile(common_file))



"""This module ensures input in the examples directory
could pass the argument checking.
"""

import json
import unittest
from pathlib import Path
from typing import Sequence, Tuple

from dargs import Argument

from dpdispatcher.arginfo import machine_dargs, resources_dargs, task_dargs

# directory of examples
p_examples = Path(__file__).parent.parent / "examples"

machine_args = machine_dargs()
resources_args = resources_dargs(detail_kwargs=False)
task_args = task_dargs()

# input_files : tuple[tuple[Argument, Path]]
#   tuple of example list
input_files: Sequence[Tuple[Argument, Path]] = (
    (machine_args, p_examples / "machine" / "expanse.json"),
    (machine_args, p_examples / "machine" / "lazy_local.json"),
    (machine_args, p_examples / "machine" / "mandu.json"),
    (resources_args, p_examples / "resources" / "expanse_cpu.json"),
    (resources_args, p_examples / "resources" / "mandu.json"),
    (resources_args, p_examples / "resources" / "tiger.json"),
    (task_args, p_examples / "task" / "deepmd-kit.json"),
    (task_args, p_examples / "task" / "g16.json"),
)


class TestExamples(unittest.TestCase):
    def test_arguments(self):
        for arginfo, fn in input_files:
            fn = str(fn)
            with self.subTest(fn=fn):
                with open(fn) as f:
                    data = json.load(f)
                arginfo.check_value(data, strict=True)



import os
import shutil
import time

from dpdispatcher.utils.utils import customized_script_header_template

try:
    from bohriumsdk.client import Client
    from bohriumsdk.job import Job
    from bohriumsdk.storage import Storage
    from bohriumsdk.util import Util
except ModuleNotFoundError:
    found_bohriumsdk = False
else:
    found_bohriumsdk = True

from dpdispatcher.dlog import dlog
from dpdispatcher.machine import Machine
from dpdispatcher.utils.job_status import JobStatus

shell_script_header_template = """
#!/bin/bash -l
"""


class OpenAPI(Machine):
    def __init__(self, context):
        if not found_bohriumsdk:
            raise ModuleNotFoundError(
                "bohriumsdk not installed. Install dpdispatcher with `pip install dpdispatcher[bohrium]`"
            )
        self.context = context
        self.remote_profile = context.remote_profile.copy()

        self.grouped = self.remote_profile.get("grouped", True)
        self.retry_count = self.remote_profile.get("retry_count", 3)
        self.ignore_exit_code = context.remote_profile.get("ignore_exit_code", True)
        self.client = Client()
        self.job = Job(client=self.client)
        self.storage = Storage(client=self.client)
        self.group_id = None

    def gen_script(self, job):
        shell_script = super().gen_script(job)
        return shell_script

    def gen_script_header(self, job):
        resources = job.resources
        if (
            resources["strategy"].get("customized_script_header_template_file")
            is not None
        ):
            shell_script_header = customized_script_header_template(
                resources["strategy"]["customized_script_header_template_file"],
                resources,
            )
        else:
            shell_script_header = shell_script_header_template
        return shell_script_header

    def gen_local_script(self, job):
        script_str = self.gen_script(job)
        script_file_name = job.script_file_name
        self.context.write_local_file(fname=script_file_name, write_str=script_str)
        script_run_str = self.gen_script_command(job)
        script_run_file_name = f"{job.script_file_name}.run"
        self.context.write_local_file(
            fname=script_run_file_name, write_str=script_run_str
        )
        return script_file_name

    def _gen_backward_files_list(self, job):
        result_file_list = []
        # result_file_list.extend(job.backward_common_files)
        for task in job.job_task_list:
            result_file_list.extend(
                [os.path.join(task.task_work_path, b_f) for b_f in task.backward_files]
            )
        result_file_list = list(set(result_file_list))
        return result_file_list

    def do_submit(self, job):
        self.gen_local_script(job)

        project_id = self.remote_profile.get("project_id", 0)

        openapi_params = {
            "oss_path": job.upload_path,
            "input_file_type": 3,
            "input_file_method": 1,
            "job_type": "container",
            "job_name": self.remote_profile.get("job_name", "DP-GEN"),
            "project_id": project_id,
            "scass_type": self.remote_profile.get("machine_type", ""),
            "cmd": f"bash {job.script_file_name}",
            "log_files": os.path.join(
                job.job_task_list[0].task_work_path, job.job_task_list[0].outlog
            ),
            "out_files": self._gen_backward_files_list(job),
            "platform": self.remote_profile.get("platform", "ali"),
            "image_address": self.remote_profile.get("image_address", ""),
        }
        if job.job_state == JobStatus.unsubmitted:
            openapi_params["job_id"] = job.job_id

        data = self.job.insert(**openapi_params)

        job.job_id = data.get("jobId", 0)  # type: ignore
        # self.job_group_id = data.get("jobGroupId")
        job.job_state = JobStatus.waiting
        return job.job_id

    def _get_job_detail(self, job_id, group_id):
        check_return = self.job.detail(job_id)
        assert check_return is not None, (
            f"Failed to retrieve tasks information. To resubmit this job, please "
            f"try again, if this problem still exists please delete the submission "
            f"file and try again.\nYou can check submission.submission_hash in the "
            f'previous log or type `grep -rl "{job_id}:job_group_id:{group_id}" '
            f"~/.dpdispatcher/dp_cloud_server/` to find corresponding file. "
            f"You can try with command:\n    "
            f'rm $(grep -rl "{job_id}:job_group_id:{group_id}" ~/.dpdispatcher/dp_cloud_server/)'
        )
        return check_return

    def check_status(self, job):
        if job.job_id == "":
            return JobStatus.unsubmitted
        job_id = job.job_id
        group_id = None
        if hasattr(job, "jgid"):
            group_id = job.jgid
        check_return = self._get_job_detail(job_id, group_id)
        try:
            dp_job_status = check_return["status"]  # type: ignore
        except IndexError as e:
            dlog.error(
                f"cannot find job information in bohrium for job {job.job_id}. check_return:{check_return}; retry one more time after 60 seconds"
            )
            time.sleep(60)
            retry_return = self._get_job_detail(job_id, group_id)
            try:
                dp_job_status = retry_return["status"]  # type: ignore
            except IndexError as e:
                raise RuntimeError(
                    f"cannot find job information in bohrium for job {job.job_id} {check_return} {retry_return}"
                )

        job_state = self.map_dp_job_state(
            dp_job_status,
            check_return.get("exitCode", 0),  # type: ignore
            self.ignore_exit_code,
        )
        if job_state == JobStatus.finished:
            job_log = self.job.log(job_id)
            if self.remote_profile.get("output_log"):
                print(job_log, end="")
            self._download_job(job)
        elif self.remote_profile.get("output_log") and job_state == JobStatus.running:
            job_log = self.job.log(job_id)
            print(job_log, end="")
        return job_state

    def _download_job(self, job):
        data = self.job.detail(job.job_id)
        job_url = data["jobFiles"]["outFiles"][0]["url"]  # type: ignore
        if not job_url:
            return
        job_hash = job.job_hash
        result_filename = job_hash + "_back.zip"
        target_result_zip = os.path.join(self.context.local_root, result_filename)
        self.storage.download_from_url(job_url, target_result_zip)
        Util.unzip_file(target_result_zip, out_dir=self.context.local_root)
        try:
            os.makedirs(os.path.join(self.context.local_root, "backup"), exist_ok=True)
            shutil.move(
                target_result_zip,
                os.path.join(
                    self.context.local_root,
                    "backup",
                    os.path.split(target_result_zip)[1],
                ),
            )
        except (OSError, shutil.Error) as e:
            dlog.exception("unable to backup file, " + str(e))

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        dlog.info("check if job finished: ", job.job_id, job_tag_finished)
        return self.context.check_file_exists(job_tag_finished)
        # return
        # pass

    def check_if_recover(self, submission):
        return False
        # pass

    @staticmethod
    def map_dp_job_state(status, exit_code, ignore_exit_code=True):
        if isinstance(status, JobStatus):
            return status
        map_dict = {
            -1: JobStatus.terminated,
            0: JobStatus.waiting,
            1: JobStatus.running,
            2: JobStatus.finished,
            3: JobStatus.waiting,
            4: JobStatus.running,
            5: JobStatus.terminated,
            6: JobStatus.running,
            9: JobStatus.waiting,
        }
        if status not in map_dict:
            dlog.error(f"unknown job status {status}")
            return JobStatus.unknown
        if status == -1 and exit_code != 0 and ignore_exit_code:
            return JobStatus.finished
        return map_dict[status]

    def kill(self, job):
        """Kill the job.

        Parameters
        ----------
        job : Job
            job
        """
        job_id = job.job_id
        self.job.kill(job_id)

    def get_exit_code(self, job):
        """Get exit code of the job.

        Parameters
        ----------
        job : Job
            job

        Returns
        -------
        int
            exit code
        """
        check_return = self.job.detail(job.job_id)
        return check_return.get("exitCode", -999)  # type: ignore



import os
import sys
import textwrap
import unittest
from pathlib import Path

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

from test_run_submission import RunSubmission


@unittest.skipIf(
    os.environ.get("DPDISPATCHER_TEST") != "bohrium",
    "outside the Bohrium testing environment",
)
class TestBohriumRun(RunSubmission, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.machine_dict.update(
            batch_type="Bohrium",
            context_type="Bohrium",
            remote_profile={
                "email": os.environ["BOHRIUM_EMAIL"],
                "password": os.environ["BOHRIUM_PASSWORD"],
                "project_id": int(os.environ["BOHRIUM_PROJECT_ID"]),
                "input_data": {
                    "job_type": "indicate",
                    "log_file": "log",
                    "job_name": "dpdispather_test",
                    "disk_size": 20,
                    "scass_type": "c2_m4_cpu",
                    "platform": "ali",
                    "image_name": "registry.dp.tech/dptech/ubuntu:22.04-py3.10",
                    "on_demand": 0,
                },
            },
        )

    @unittest.skip("Manaually skip")  # comment this line to open unittest
    def test_async_run_submission(self):
        return super().test_async_run_submission()


@unittest.skipIf(
    os.environ.get("DPDISPATCHER_TEST") != "bohrium",
    "outside the Bohrium testing environment",
)
class TestOpenAPIRun(RunSubmission, unittest.TestCase):
    def setUp(self):
        super().setUp()
        bohrium_config = textwrap.dedent(
            """\
            [Credentials]
            accessKey={accesskey}
            """
        ).format(accesskey=os.environ["BOHRIUM_ACCESS_KEY"])
        Path.home().joinpath(".brmconfig").write_text(bohrium_config)
        self.machine_dict.update(
            batch_type="OpenAPI",
            context_type="OpenAPI",
            remote_profile={
                "project_id": int(os.environ["BOHRIUM_PROJECT_ID"]),
                "machine_type": "c2_m4_cpu",
                "platform": "ali",
                "image_address": "registry.dp.tech/dptech/ubuntu:22.04-py3.10",
                "job_name": "dpdispather_test",
            },
        )

    @unittest.skip("Manaually skip")  # comment this line to open unittest
    def test_async_run_submission(self):
        return super().test_async_run_submission()



from pathlib import Path

from dpdispatcher.dlog import dlog
from dpdispatcher.submission import Submission
from dpdispatcher.utils.job_status import JobStatus
from dpdispatcher.utils.record import record


def handle_submission(
    *,
    submission_hash: str,
    download_terminated_log: bool = False,
    download_finished_task: bool = False,
    clean: bool = False,
    reset_fail_count: bool = False,
):
    """Handle terminated submission.

    Parameters
    ----------
    submission_hash : str
        Submission hash to download.
    download_terminated_log : bool, optional
        Download log files of terminated tasks.
    download_finished_task : bool, optional
        Download finished tasks.
    clean : bool, optional
        Clean submission.
    reset_fail_count : bool, optional
        Reset fail count of all jobs to zero.

    Raises
    ------
    ValueError
        At least one action should be specified.
    """
    if (
        int(download_terminated_log)
        + int(download_finished_task)
        + int(clean)
        + int(reset_fail_count)
        == 0
    ):
        raise ValueError("At least one action should be specified.")

    submission_file = record.get_submission(submission_hash)
    submission = Submission.submission_from_json(str(submission_file))
    submission.belonging_tasks = [
        task for job in submission.belonging_jobs for task in job.job_task_list
    ]
    # TODO: for unclear reason, the submission_hash may be changed
    submission.submission_hash = submission_hash
    submission.machine.context.bind_submission(submission)
    if reset_fail_count:
        for job in submission.belonging_jobs:
            job.fail_count = 0
        # save to remote and local
        submission.submission_to_json()
        record.write(submission)
    if int(download_terminated_log) + int(download_finished_task) + int(clean) == 0:
        # if only reset_fail_count, no need to update submission state (expensive)
        return
    submission.update_submission_state()
    submission.submission_to_json()
    record.write(submission)

    terminated_tasks = []
    finished_tasks = []
    for task in submission.belonging_tasks:
        task.get_task_state(submission.machine.context)
        if task.task_state == JobStatus.terminated:
            terminated_tasks.append(task)
        elif task.task_state == JobStatus.finished:
            finished_tasks.append(task)
    submission.belonging_tasks = []

    if download_terminated_log:
        for task in terminated_tasks:
            task.backward_files = [task.outlog, task.errlog]
        submission.belonging_tasks += terminated_tasks
    if download_finished_task:
        submission.belonging_tasks += finished_tasks

    submission.download_jobs()

    if download_terminated_log:
        terminated_log_files = []
        for task in terminated_tasks:
            assert submission.local_root is not None
            terminated_log_files.append(
                Path(submission.local_root) / task.task_work_path / task.outlog
            )
            terminated_log_files.append(
                Path(submission.local_root) / task.task_work_path / task.errlog
            )

        dlog.info(
            "Terminated logs are downloaded into:\n  "
            + "\n  ".join([str(f) for f in terminated_log_files])
        )

    if clean:
        submission.clean_jobs()



__author__ = "DeepModeling Team"
__copyright__ = "Copyright 2019-2023, DeepModeling"
__status__ = "Production"
try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

import dpdispatcher.contexts  # noqa: F401
import dpdispatcher.machines  # noqa: F401
from dpdispatcher.machine import Machine
from dpdispatcher.submission import Job, Resources, Submission, Task

__all__ = [
    "__version__",
    "Machine",
    "Submission",
    "Task",
    "Job",
    "Resources",
]



# 2开头的错误代码第二位代表错误等级
# 0. 严重错误; 1. 普通错误; 2. 规则错误; 3. 一般信息; 4. 未知错误
class RETCODE:
    OK = "0000"  # 正常
    DBERR = "2000"  # 数据库异常
    THIRDERR = "2001"  # 第三方异常
    DATAERR = "2002"  # 数据异常
    IOERR = "2003"  # IO异常

    TOKENINVALID = "2100"  # 登陆错误
    PARAMERR = "2101"  # 参数错误
    USERERR = "2102"  # 用户异常
    ROLEERR = "2103"  # 权限错误
    PWDERR = "2104"  # 密码错误
    VERIFYERR = "2105"  # 验证错误

    REQERR = "2200"  # 请求错误

    NODATA = "2300"  # 无数据
    UNDERDEBUG = "2301"  # debug模式下无法使用

    UNKOWNERR = "2400"  # 未知错误



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from datetime import date

# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "DPDispatcher"
copyright = "2020-%d, Deep Modeling" % date.today().year
author = "DeepModeling"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "deepmodeling_sphinx",
    "dargs.sphinx",
    "myst_parser",
    "sphinx_rtd_theme",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "numpydoc",
    "sphinx.ext.autosummary",
    "sphinxarg.ext",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = []

autodoc_default_flags = ["members"]
autosummary_generate = True
master_doc = "index"


def run_apidoc(_):
    from sphinx.ext.apidoc import main

    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    module = os.path.join(cur_dir, "..", "dpdispatcher")
    main(
        [
            "-M",
            "--tocfile",
            "api",
            "-H",
            "DPDispatcher API",
            "-o",
            os.path.join(cur_dir, "api"),
            module,
            "--force",
        ]
    )


def setup(app):
    app.connect("builder-inited", run_apidoc)


intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "dargs": ("https://docs.deepmodeling.com/projects/dargs/en/latest/", None),
}

myst_enable_extensions = [
    "colon_fence",
]



import shlex

from dpdispatcher.dlog import dlog
from dpdispatcher.machine import Machine
from dpdispatcher.utils.job_status import JobStatus
from dpdispatcher.utils.utils import customized_script_header_template

fugaku_script_header_template = """\
{queue_name_line}
{fugaku_node_number_line}
{fugaku_ntasks_per_node_line}
"""


class Fugaku(Machine):
    def gen_script(self, job):
        fugaku_script = super().gen_script(job)
        return fugaku_script

    def gen_script_header(self, job):
        resources = job.resources
        fugaku_script_header_dict = {}
        fugaku_script_header_dict["fugaku_node_number_line"] = (
            f'#PJM -L "node={resources.number_node}" '
        )
        fugaku_script_header_dict["fugaku_ntasks_per_node_line"] = (
            f'#PJM --mpi "max-proc-per-node={resources.cpu_per_node}"'
        )
        fugaku_script_header_dict["queue_name_line"] = (
            f'#PJM -L "rscgrp={resources.queue_name}"'
        )
        if (
            resources["strategy"].get("customized_script_header_template_file")
            is not None
        ):
            fugaku_script_header = customized_script_header_template(
                resources["strategy"]["customized_script_header_template_file"],
                resources,
            )
        else:
            fugaku_script_header = fugaku_script_header_template.format(
                **fugaku_script_header_dict
            )
        return fugaku_script_header

    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + "_job_id"
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        self.context.write_file(fname=script_file_name, write_str=script_str)
        script_run_str = self.gen_script_command(job)
        script_run_file_name = f"{job.script_file_name}.run"
        self.context.write_file(fname=script_run_file_name, write_str=script_run_str)
        # self.context.write_file(fname=os.path.join(self.context.submission.work_base, script_file_name), write_str=script_str)
        # script_file_dir = os.path.join(self.context.submission.work_base)
        script_file_dir = self.context.remote_root
        # stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (self.context.remote_root, 'pjsub', script_file_name))

        stdin, stdout, stderr = self.context.block_checkcall(
            "cd {} && {} {}".format(
                shlex.quote(script_file_dir), "pjsub", shlex.quote(script_file_name)
            )
        )
        subret = stdout.readlines()
        job_id = subret[0].split()[5]
        self.context.write_file(job_id_name, job_id)
        return job_id

    def default_resources(self, resources):
        pass

    def check_status(self, job):
        job_id = job.job_id
        if job_id == "":
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr = self.context.block_call("pjstat " + job_id)
        err_str = stderr.read().decode("utf-8")
        try:
            status_line = stdout.read().decode("utf-8").split("\n")[-2]
        # pjstat only retrun 0 if the job is not waiting or running
        except Exception:
            ret, stdin, stdout, stderr = self.context.block_call("pjstat -H  " + job_id)
            status_line = stdout.read().decode("utf-8").split("\n")[-2]
            status_word = status_line.split()[3]
            if status_word in ["EXT", "CCL", "ERR"]:
                if self.check_finish_tag(job):
                    dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                    return JobStatus.finished
                else:
                    return JobStatus.terminated
            else:
                return JobStatus.unknown
        status_word = status_line.split()[3]
        # dlog.info (status_word)
        if status_word in ["QUE", "HLD", "RNA", "SPD"]:
            return JobStatus.waiting
        elif status_word in ["RUN", "RNE"]:
            return JobStatus.running
        else:
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        return self.context.check_file_exists(job_tag_finished)



"""Machines."""

import importlib
from pathlib import Path

PACKAGE_BASE = "dpdispatcher.machines"
NOT_LOADABLE = ("__init__.py",)

for module_file in Path(__file__).parent.glob("*.py"):
    if module_file.name not in NOT_LOADABLE:
        module_name = f".{module_file.stem}"
        importlib.import_module(module_name, PACKAGE_BASE)



import os
import sys
import tempfile
import unittest
from pathlib import Path

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"

from .context import run


class TestRun(unittest.TestCase):
    def test_run(self):
        this_dir = Path(__file__).parent
        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as temp_dir:
            try:
                os.chdir(temp_dir)
                run(filename=str(this_dir / "hello_world.py"))
                self.assertEqual(
                    (Path(temp_dir) / "log").read_text().strip(), "hello world!"
                )
            finally:
                os.chdir(cwd)



import shlex

from dpdispatcher.dlog import dlog
from dpdispatcher.machine import Machine
from dpdispatcher.utils.job_status import JobStatus
from dpdispatcher.utils.utils import customized_script_header_template

shell_script_header_template = """
#!/bin/bash -l
"""


class Shell(Machine):
    def gen_script(self, job):
        shell_script = super().gen_script(job)
        return shell_script

    def gen_script_header(self, job):
        resources = job.resources
        if (
            resources["strategy"].get("customized_script_header_template_file")
            is not None
        ):
            shell_script_header = customized_script_header_template(
                resources["strategy"]["customized_script_header_template_file"],
                resources,
            )
        else:
            shell_script_header = shell_script_header_template
        return shell_script_header

    def do_submit(self, job):
        script_str = self.gen_script(job)
        script_file_name = job.script_file_name
        job_id_name = job.job_hash + "_job_id"
        output_name = job.job_hash + ".out"
        self.context.write_file(fname=script_file_name, write_str=script_str)
        script_run_str = self.gen_script_command(job)
        script_run_file_name = f"{job.script_file_name}.run"
        self.context.write_file(fname=script_run_file_name, write_str=script_run_str)
        cmd = f"cd {shlex.quote(self.context.remote_root)} && {{ nohup bash {script_file_name} 1>>{output_name} 2>>{output_name} & }} && echo $!"
        ret, stdin, stdout, stderr = self.context.block_call(cmd)
        if ret != 0:
            err_str = stderr.read().decode("utf-8")
            raise RuntimeError(
                "status command %s fails to execute\nerror message:%s\nreturn code %d\n"
                % (cmd, err_str, ret)
            )
        job_id = int(stdout.read().decode("utf-8").strip())
        self.context.write_file(job_id_name, str(job_id))
        return job_id

        # script_file_name = job.script_file_name
        # script_str = self.gen_script(job)
        # job_id_name = job.job_hash + '_job_id'
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        # self.context.write_file(fname=script_file_name, write_str=script_str)
        # stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (self.context.remote_root, 'qsub', script_file_name))
        # subret = (stdout.readlines())
        # job_id = subret[0].split()[0]
        # self.context.write_file(job_id_name, job_id)
        # return job_id

    def default_resources(self, resources):
        pass

    def check_status(self, job):
        job_id = job.job_id
        # print('shell.check_status.job_id', job_id)
        # job_state = JobStatus.unknown
        if job_id == "":
            return JobStatus.unsubmitted

        # mark defunct process as terminated
        cmd = (
            r"""command -v ps >/dev/null 2>&1 || { echo >&2 "I require ps but it's not installed. Aborting."; exit 1; };"""
            f"if ps -p {job_id} > /dev/null && ! (ps -o command -p {job_id} | grep defunct >/dev/null) ; then echo 1; fi"
        )
        ret, stdin, stdout, stderr = self.context.block_call(cmd)
        if ret != 0:
            err_str = stderr.read().decode("utf-8")
            raise RuntimeError(
                "status command %s fails to execute\nerror message:%s\nreturn code %d\n"
                % (cmd, err_str, ret)
            )

        if_job_exists = bool(stdout.read().decode("utf-8").strip())
        if self.check_finish_tag(job=job):
            dlog.info(f"job: {job.job_hash} {job.job_id} finished")
            return JobStatus.finished

        if if_job_exists:
            return JobStatus.running
        else:
            return JobStatus.terminated
        # return job_state

    # def check_status(self, job):
    #     job_id = job.job_id
    #     uuid_names = job.job_hash
    #     cnt = 0
    #     ret, stdin, stdout, stderr = self.context.block_call("ps aux | grep %s"%uuid_names)
    #     response_list = stdout.read().decode('utf-8').split("\n")
    #     for response in response_list:
    #         if  uuid_names + ".sub" in response:
    #             return True
    #     return False

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        # print('job finished: ',job.job_id, job_tag_finished)
        return self.context.check_file_exists(job_tag_finished)

    def kill(self, job):
        """Kill the job.

        Parameters
        ----------
        job : Job
            job
        """
        job_id = job.job_id
        # 9 means exit, cannot be blocked
        ret, stdin, stdout, stderr = self.context.block_call("kill -9 " + str(job_id))



# SPDX-License-Identifier: LGPL-3.0-or-later
import unittest

from dpgui import (
    generate_dpgui_templates,
)


class TestDPGUI(unittest.TestCase):
    def test_dpgui_entrypoints(self):
        self.assertTrue(len(generate_dpgui_templates()) > 0)



import shlex
from typing import List

from dargs import Argument

from dpdispatcher.dlog import dlog
from dpdispatcher.machine import Machine
from dpdispatcher.utils.job_status import JobStatus
from dpdispatcher.utils.utils import (
    RetrySignal,
    customized_script_header_template,
    retry,
)

JH_UniScheduler_script_header_template = """\
#!/bin/bash -l
#JSUB -e %J.err
#JSUB -o %J.out
{JH_UniScheduler_nodes_line}
{JH_UniScheduler_ptile_line}
{JH_UniScheduler_partition_line}
{JH_UniScheduler_number_gpu_line}"""


class JH_UniScheduler(Machine):
    """JH_UniScheduler batch."""

    def gen_script(self, job):
        JH_UniScheduler_script = super().gen_script(job)
        return JH_UniScheduler_script

    def gen_script_header(self, job):
        resources = job.resources
        script_header_dict = {
            "JH_UniScheduler_nodes_line": f"#JSUB -n {resources.number_node * resources.cpu_per_node}",
            "JH_UniScheduler_ptile_line": f"#JSUB -R 'span[ptile={resources.cpu_per_node}]'",
            "JH_UniScheduler_partition_line": f"#JSUB -q {resources.queue_name}",
        }
        custom_gpu_line = resources.kwargs.get("custom_gpu_line", None)
        if not custom_gpu_line:
            script_header_dict["JH_UniScheduler_number_gpu_line"] = (
                "" f"#JSUB -gpgpu {resources.gpu_per_node}"
            )
        else:
            script_header_dict["JH_UniScheduler_number_gpu_line"] = custom_gpu_line
        if (
            resources["strategy"].get("customized_script_header_template_file")
            is not None
        ):
            JH_UniScheduler_script_header = customized_script_header_template(
                resources["strategy"]["customized_script_header_template_file"],
                resources,
            )
        else:
            JH_UniScheduler_script_header = (
                JH_UniScheduler_script_header_template.format(**script_header_dict)
            )

        return JH_UniScheduler_script_header

    @retry()
    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + "_job_id"
        self.context.write_file(fname=script_file_name, write_str=script_str)
        script_run_str = self.gen_script_command(job)
        script_run_file_name = f"{job.script_file_name}.run"
        self.context.write_file(fname=script_run_file_name, write_str=script_run_str)

        try:
            stdin, stdout, stderr = self.context.block_checkcall(
                "cd {} && {} {}".format(
                    shlex.quote(self.context.remote_root),
                    "jsub < ",
                    shlex.quote(script_file_name),
                )
            )
        except RuntimeError as err:
            raise RetrySignal(err) from err

        subret = stdout.readlines()
        job_id = subret[0].split()[1][1:-1]
        self.context.write_file(job_id_name, job_id)
        return job_id

    def default_resources(self, resources):
        pass

    @retry()
    def check_status(self, job):
        try:
            job_id = job.job_id
        except AttributeError:
            return JobStatus.terminated
        if job_id == "":
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr = self.context.block_call("jjobs " + job_id)
        err_str = stderr.read().decode("utf-8")
        if (f"Job <{job_id}> is not found") in err_str:
            if self.check_finish_tag(job):
                return JobStatus.finished
            else:
                return JobStatus.terminated
        elif ret != 0:
            # just retry when any unknown error raised.
            raise RetrySignal(
                "Get error code %d in checking status with job: %s . message: %s"
                % (ret, job.job_hash, err_str)
            )
        status_out = stdout.read().decode("utf-8").split("\n")
        if len(status_out) < 2:
            return JobStatus.unknown
        else:
            status_line = status_out[1]
            status_word = status_line.split()[2]

        if status_word in ["PEND"]:
            return JobStatus.waiting
        elif status_word in ["RUN", "PSUSP", "SSUSP", "USUSP"]:
            return JobStatus.running
        elif status_word in ["DONE", "EXIT"]:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else:
                return JobStatus.terminated
        else:
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        return self.context.check_file_exists(job_tag_finished)

    @classmethod
    def resources_subfields(cls) -> List[Argument]:
        """Generate the resources subfields.

        Returns
        -------
        list[Argument]
            resources subfields
        """
        doc_custom_gpu_line = "Custom GPU configuration, starting with #JSUB"

        return [
            Argument(
                "kwargs",
                dict,
                [
                    Argument(
                        "custom_gpu_line",
                        str,
                        optional=True,
                        default=None,
                        doc=doc_custom_gpu_line,
                    ),
                ],
                optional=False,
                doc="Extra arguments.",
            )
        ]

    def kill(self, job):
        """Kill the job.

        Parameters
        ----------
        job : Job
            job
        """
        job_id = job.job_id
        ret, stdin, stdout, stderr = self.context.block_call(
            "jctrl kill " + str(job_id)
        )



# import os,sys,json,glob,shutil,uuid,time
# import unittest
# from unittest.mock import MagicMock, patch, PropertyMock

# sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
# __package__ = 'tests'
# from .context import LocalSession
# from .context import LocalContext
# from .context import PBS
# from .context import JobStatus
# from .context import Submission, Job, Task, Resources
from sample_class import SampleClass

# print('in', SampleClass.get_sample_empty_submission())


empty_submission = SampleClass.get_sample_empty_submission()
task = SampleClass.get_sample_task()
# print('TestSubmissionInit.test_reigister_task:self.empty_submission.belonging_tasks', empty_submission.belonging_tasks)
empty_submission.register_task(task=task)
# print('7890809', SampleClass.get_sample_empty_submission().belonging_tasks)
empty_submission.register_task(task=task)
# print('1441198', SampleClass.get_sample_empty_submission().belonging_tasks)
# self.assertEqual([task], empty_submission.belonging_tasks)
# print('out', SampleClass.get_sample_empty_submission())
# print('TestSubmissionInit.test_register_task_list:task_list', task_list)
# empty_submission = SampleClass.get_sample_empty_submission()
# task_list = SampleClass.get_sample_task_list()
# empty_submission.register_task_list(task_list=task_list)
# self.empty_submission.register_task_list(task_list=task_list)
# self.assertEqual(task_list, empty_submission.belonging_tasks)

# def tesk_generate_jobs(self):
#     task_list = SampleClass.get_sample_task_list()
#     self.submission.register_task_list(task_list=task_list)
#     self.submission.generate_jobs()
#     task1, task2, task3, task4 = task_list
#     task_ll = [job.job_task_list for job in self.submission.belonging_jobs]
#     self.assertEqual([[task3, task2], [task4, task1]], task_ll)



import os
import sys
import unittest
from socket import gaierror

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
from dargs.dargs import ArgumentValueError

from .context import (  # noqa: F401
    LSF,
    PBS,
    BaseContext,
    DistributedShell,
    JH_UniScheduler,
    LazyLocalContext,
    Lebesgue,
    LocalContext,
    Machine,
    Shell,
    Slurm,
    setUpModule,
)


class TestMachineDispatch(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_init_machine_pbs_lazy_local(self):
        machine_dict = {
            "batch_type": "PBS",
            "context_type": "LazyLocalContext",
            "local_root": "./",
        }

        machine = Machine(**machine_dict)
        self.assertIsInstance(machine, PBS)
        self.assertIsInstance(machine.context, LazyLocalContext)

    def test_init_machine_shell_local(self):
        machine_dict = {
            "batch_type": "Shell",
            "context_type": "LocalContext",
            "local_root": "./",
            "remote_root": "/tmp",
        }

        shell = Shell(**machine_dict)
        self.assertIsInstance(shell, Shell)
        self.assertIsInstance(shell.context, LocalContext)

    def test_init_machine_slurm_ssh(self):
        machine_dict = {
            "batch_type": "Slurm",
            "context_type": "SSHContext",
            "local_root": "./",
            "remote_root": "/tmp/dpdispatcher",
            "remote_profile": {"hostname": "39.106.xx.xx", "username": "test1234"},
        }

        with self.assertRaises(gaierror):
            Machine(**machine_dict)

    def test_lazy_local(self):
        machine_dict = {
            "batch_type": "PBS",
            "context_type": "LazyLocalContext",
            "local_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        # pylint: disable=maybe-no-member
        self.assertIsInstance(machine.context, LazyLocalContext)

    def test_lower_case(self):
        machine_dict = {
            "batch_type": "pbs",
            "context_type": "lazylocalcontext",
            "local_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine.context, LazyLocalContext)

    def test_no_ending_context(self):
        machine_dict = {
            "batch_type": "PBS",
            "context_type": "lazylocal",
            "local_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine.context, LazyLocalContext)

    def test_local(self):
        machine_dict = {
            "batch_type": "PBS",
            "context_type": "LocalContext",
            "local_root": "./",
            "remote_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        # pylint: disable=maybe-no-member
        self.assertIsInstance(machine.context, LocalContext)

    def test_ssh(self):
        pass
        # jdata = {
        #     'batch_type': 'pbs',
        #     'context_type': 'ssh',
        #     'hostname': 'localhost',
        #     'local_root': './',
        #     "remote_root" : "/home/fengbo/work_path_dpdispatcher_test/",
        #     "username" : "fengbo"
        # }
        # batch = BatchObject(
        #     jdata=jdata
        # )
        # self.assertIsInstance(batch.context, SSHContext)

    def test_key_err(self):
        # pass
        machine_dict = {}
        with self.assertRaises(KeyError):
            Machine.load_from_dict(machine_dict=machine_dict)

    def test_context_err(self):
        machine_dict = {"batch_type": "PBS", "context_type": "foo"}
        # with self.assertRaises(KeyError):
        with self.assertRaises(ArgumentValueError):
            Machine.load_from_dict(machine_dict=machine_dict)

    def test_pbs(self):
        machine_dict = {
            "batch_type": "PBS",
            "context_type": "LazyLocalContext",
            "local_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine, PBS)

    def test_lsf(self):
        machine_dict = {
            "batch_type": "LSF",
            "context_type": "LazyLocalContext",
            "local_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine, LSF)

    def test_slurm(self):
        machine_dict = {
            "batch_type": "Slurm",
            "context_type": "LazyLocalContext",
            "local_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine, Slurm)

    def test_jh_unischeduler(self):
        machine_dict = {
            "batch_type": "JH_UniScheduler",
            "context_type": "LazyLocalContext",
            "local_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine, JH_UniScheduler)

    def test_shell(self):
        machine_dict = {
            "batch_type": "Shell",
            "context_type": "LazyLocalContext",
            "local_root": "./",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine, Shell)

    def test_distributed_shell(self):
        machine_dict = {
            "batch_type": "DistributedShell",
            "context_type": "HDFSContext",
            "local_root": "./",
            "remote_root": "hdfs://test/",
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine, DistributedShell)

    def test_lebesgue(self):
        machine_dict = {
            "batch_type": "Lebesgue",
            "context_type": "LebesgueContext",
            "local_root": "./",
            "remote_root": "./",
            "remote_profile": {
                "email": "114@514.com",
                "password": "114514",
                "program_id": 114514,
                "input_data": {},
            },
        }
        machine = Machine.load_from_dict(machine_dict=machine_dict)
        self.assertIsInstance(machine, Lebesgue)


class TestContextDispatch(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_init_lazy_local(self):
        context_dict = {
            "context_type": "LazyLocalContext",
            "local_root": "./",
        }

        context = BaseContext(**context_dict)
        self.assertIsInstance(context, LazyLocalContext)

    def test_subclass_init_local(self):
        context_dict = {
            "context_type": "LocalContext",
            "local_root": "./",
            "remote_root": "/tmp",
        }

        context = LocalContext(**context_dict)
        self.assertIsInstance(context, LocalContext)

    def test_init_local(self):
        context_dict = {
            "context_type": "LocalContext",
            "local_root": "./",
            "remote_root": "/tmp",
        }

        context = BaseContext(**context_dict)
        self.assertIsInstance(context, LocalContext)



import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from dpdispatcher.local_context import LocalContext, LocalSession

# from dpdispatcher.pbs import PBS
from dpdispatcher.shell import Shell
from dpdispatcher.submission import Resources, Submission, Task

local_session = LocalSession({"work_path": "temp2"})
local_context = LocalContext(local_root="test_shell_dir/", work_profile=local_session)
# lazy_local_context = LazyLocalContext(local_root='/home/fengbo/10_dpdispatcher/dpdispatcher/tests/temp3/0_md', work_profile=None)
shell = Shell(context=local_context)
# pbs = PBS(context=lazy_local_context)

resources = Resources(
    number_node=1, cpu_per_node=4, gpu_per_node=1, queue_name="V100_8_32", group_size=4
)
submission = Submission(
    work_base="0_md",
    resources=resources,
    forward_common_files=["graph.pb"],
    backward_common_files=["submission.json"],
)  # ,  batch=PBS)
task1 = Task(
    command="lmp_serial -i input.lammps",
    task_work_path="bct-1",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
    task_need_resources=1,
)
task2 = Task(
    command="lmp_serial -i input.lammps",
    task_work_path="bct-2",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
    task_need_resources=0.25,
)
task3 = Task(
    command="lmp_serial -i input.lammps",
    task_work_path="bct-3",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
    task_need_resources=0.25,
)
task4 = Task(
    command="lmp_serial -i input.lammps",
    task_work_path="bct-4",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
    task_need_resources=0.5,
)
submission.register_task_list(
    [
        task1,
        task2,
        task3,
        task4,
    ]
)
submission.generate_jobs()
submission.bind_batch(batch=shell)
# for job in submission.belonging_jobs:
#     job.job_to_json()
# print('111', submission)
# submission2 = Submission.recover_jobs_from_json('./jr.json')
# print('222', submission2)
# print(submission==submission2)
submission.run_submission()

# submission1.dump_jobs_fo_json()
# submission2 = Submission.submission_from_json('jsons/submission.json')
# print(677, submission==submission2)
# print(submission1.belonging_jobs)
# print(local_context)



import os
import re
import sys
from glob import glob
from hashlib import sha1

from dpdispatcher.machine import Machine
from dpdispatcher.submission import Resources, Submission, Task

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib
from typing import List, Optional

from dargs import Argument

from dpdispatcher.arginfo import machine_dargs, resources_dargs, task_dargs

REGEX = r"(?m)^# /// (?P<type>[a-zA-Z0-9-]+)$\s(?P<content>(^#(| .*)$\s)+)^# ///$"


def read_pep723(script: str) -> Optional[dict]:
    """Read a PEP 723 script metadata from a script string.

    Parameters
    ----------
    script : str
        Script content.

    Returns
    -------
    dict
        PEP 723 metadata.
    """
    name = "script"
    matches = list(
        filter(lambda m: m.group("type") == name, re.finditer(REGEX, script))
    )
    if len(matches) > 1:
        # TODO: Add tests for scenarios where multiple script blocks are found
        raise ValueError(f"Multiple {name} blocks found")
    elif len(matches) == 1:
        content = "".join(
            line[2:] if line.startswith("# ") else line[1:]
            for line in matches[0].group("content").splitlines(keepends=True)
        )
        return tomllib.loads(content)
    else:
        # TODO: Add tests for scenarios where no metadata is found
        return None


def pep723_args() -> Argument:
    """Return the argument parser for PEP 723 metadata."""
    machine_args = machine_dargs()
    machine_args.fold_subdoc = True
    machine_args.doc = "Machine configuration. See related documentation for details."
    resources_args = resources_dargs(detail_kwargs=False)
    resources_args.fold_subdoc = True
    resources_args.doc = (
        "Resources configuration. See related documentation for details."
    )
    task_args = task_dargs()
    command_arg = task_args["command"]
    command_arg.doc = (
        "Python interpreter or launcher. No need to contain the Python script filename."
    )
    command_arg.default = "python"
    command_arg.optional = True
    task_args["task_work_path"].doc += " Can be a glob pattern."
    task_args.name = "task_list"
    task_args.doc = "List of tasks to execute."
    task_args.repeat = True
    task_args.dtype = (list,)
    return Argument(
        "pep723",
        dtype=dict,
        doc="PEP 723 metadata",
        sub_fields=[
            Argument(
                "work_base",
                dtype=str,
                optional=True,
                default="./",
                doc="Base directory for the work",
            ),
            Argument(
                "forward_common_files",
                dtype=List[str],
                optional=True,
                default=[],
                doc="Common files to forward to the remote machine",
            ),
            Argument(
                "backward_common_files",
                dtype=List[str],
                optional=True,
                default=[],
                doc="Common files to backward from the remote machine",
            ),
            machine_args,
            resources_args,
            task_args,
        ],
    )


def create_submission(metadata: dict, hash: str) -> Submission:
    """Create a Submission instance from a PEP 723 metadata.

    Parameters
    ----------
    metadata : dict
        PEP 723 metadata.
    hash : str
        Submission hash.

    Returns
    -------
    Submission
        Submission instance.
    """
    base = pep723_args()
    metadata = base.normalize_value(metadata, trim_pattern="_*")
    base.check_value(metadata, strict=False)

    tasks = []
    for task in metadata["task_list"]:
        task = task.copy()
        task["command"] += f" $REMOTE_ROOT/script_{hash}.py"
        task_work_path = os.path.join(
            metadata["machine"]["local_root"],
            metadata["work_base"],
            task["task_work_path"],
        )
        if os.path.isdir(task_work_path):
            tasks.append(Task.load_from_dict(task))
        elif glob(task_work_path):
            for file in glob(task_work_path):
                tasks.append(Task.load_from_dict({**task, "task_work_path": file}))
        # TODO: Add tests for scenarios where the task work path is a glob pattern
        else:
            # TODO: Add tests for scenarios where the task work path is not found
            raise FileNotFoundError(f"Task work path {task_work_path} not found.")
    return Submission(
        work_base=metadata["work_base"],
        forward_common_files=metadata["forward_common_files"],
        backward_common_files=metadata["backward_common_files"],
        machine=Machine.load_from_dict(metadata["machine"]),
        resources=Resources.load_from_dict(metadata["resources"]),
        task_list=tasks,
    )


def run_pep723(script: str):
    """Run a PEP 723 script.

    Parameters
    ----------
    script : str
        Script content.
    """
    metadata = read_pep723(script)
    if metadata is None:
        raise ValueError("No PEP 723 metadata found.")
    dpdispatcher_metadata = metadata["tool"]["dpdispatcher"]
    script_hash = sha1(script.encode("utf-8")).hexdigest()
    submission = create_submission(dpdispatcher_metadata, script_hash)
    submission.machine.context.write_file(f"script_{script_hash}.py", script)
    # write script
    submission.run_submission()



import json
from pathlib import Path
from typing import List


class Record:
    """Record failed or canceled submissions."""

    def __init__(self) -> None:
        self.record_directory = Path.home() / ".dpdispatcher" / "submission"
        self.record_directory.mkdir(parents=True, exist_ok=True)

    def get_submissions(self) -> List[str]:
        """Get all stored submission hashes.

        Returns
        -------
        list[str]
            List of submission hashes.
        """
        return [
            f.stem
            for f in self.record_directory.iterdir()
            if (f.is_file() and f.suffix == ".json")
        ]

    def write(self, submission) -> Path:
        """Write submission data to file.

        Parameters
        ----------
        submission : dpdispatcher.Submission
            Submission data.

        Returns
        -------
        pathlib.Path
            Path to submission data.
        """
        submission_path = self.record_directory / f"{submission.submission_hash}.json"
        submission_path.write_text(json.dumps(submission.serialize(), indent=2))
        return submission_path

    def get_submission(self, hash: str, not_exist_ok: bool = False) -> Path:
        """Get submission data by hash.

        Parameters
        ----------
        hash : str
            Hash of submission data.

        Returns
        -------
        pathlib.Path
            Path to submission data.
        """
        submission_file = self.record_directory / f"{hash}.json"
        if not not_exist_ok and not submission_file.is_file():
            raise FileNotFoundError(f"Submission file not found: {submission_file}")
        return submission_file

    def remove(self, hash: str):
        """Remove submission data by hash.

        Call this method when the remote directory is cleaned.

        Parameters
        ----------
        hash : str
            Hash of submission data.
        """
        path = self.get_submission(hash, not_exist_ok=True)
        if path.is_file():
            path.unlink()


# the record object can be globally used
record = Record()
__all__ = ["record"]
